%% T-61.2010 Datasta tietoon / T#3 / 4
% Jukka Parviainen, 21.11.2006

clear;         % poista muistista muuttujat
clc;           % puhdista komentoikkunan ruutu
close all;     % sulje kaikki ikkunat

%% c-means: kiisseli
% Mys tunnetaan (yleisemmin?) nimell "k-means".
%
% Erll TKK:n tietotekniikan peruskurssilla pyydettiin 
% opiskelijoita lausumaan sana "kiisseli" ja tallentamaan 
% se WAV-muotoisena, kts. esim.
%|http://www.cis.hut.fi/Opinnot/T-61.2010/Laskarit/Kiisseli/|
%
% Nist poimittiin 10 nytett, jotka kirjoitettiin
% putkeen yhdeksi nitiedostoksi. Kukin nyte skaalattiin
% uudestaan niin, ett maksimiarvot olivat ykksi.
%
% Yhdistetyst ninytteest laskettiin 20 ms:n vlein
% 40 ms:n pitkist ikkunoista mel-taajuuden kepstrikertoimet
% (mel-frequency cepstral coefficients, MFCC). Ylimalkainen kaava:
%
% MFCC = ifft(log10(abs(fft(x))^2)
%
% MFCC:it kytetn hyvin yleisesti piirtein 
% puheentunnistusalgoritmeissa. Niiden on osoitettu
% toimivan hyvin ja purkavan korrelaatiota. Tll kertaa laskettiin
% kertoimet 13 taajuusalueelta (0-11025 Hz); Mel-asteikolla
% matalilla taajuuksilla kaistat ovat kapeampia.
%
% Alkuperisess ninytteess on 22050 nytett a 8 bitti
% sekunnissa eli 22050 tavua. Kertoimia lasketaan 20 ms:n 
% vlein 13 kpl eli sekunnissa 50*13=650 (luvun esitystarkkuus
% 2-4 tavua?).
% 
% Reaalisista kepstrikertoimista ei voida palauttaa
% takaisin alkuperist signaalia, vaan tarkoituksena
% on signaalin analysointi.
%
% Seuraavaksi on tarkoituksena ryhmitell MFCC-kertoimet
% c-means-klusteroinnilla ja kuunnella, milt kertoimia
% vastaavien ninytteiden klusterit kuulostavat.
%
% Lasketaan mys suoraan Fourier-muunnoksen itseisarvosta
% 13 tasavlisen kaistan energiaa. Kutsutaan nit
% "short-time Fourier-transform coefficients" (STFTC).
% Tehdn c-means-klusterointi nillekin.

%% Ladataan ja kuunnellaan nitiedosto
[X, Fs] = wavread('kiisseli_10x.wav');
soundsc(X, Fs);

%% Piirretn nitiedosto
% Ylkuvaajaan aaltomuoto, x-akselilla aika sekunneissa.
% Alakuvaajaan spektogrammi, x-akselilla aika, y-akselilla taajuus

t = [1 : length(X)]/Fs;

figure(41); clf;
subplot(2,1,1);
plot(t, X);
set(gca,'XLim',[t(1) t(end)]);
title('Kiisseli x 10');
grid on;
subplot(2,1,2);
specgram(X, [], Fs);
% Matlabin uusissa versioissa:
%spectrogram(X, [], [], [], Fs, 'yaxis');

%% Ladataan valmiiksi lasketut MFCC-kertoimet
% Mel-kepstikertoimet laskettu netist lytyvll
% Matlab-toolboxilla:
% |http://labrosa.ee.columbia.edu/matlab/rastamat/|
% "PLP and RASTA (and MFCC, and inversion) in Matlab".
% Kytetty komentoa:
% [mfcc] = melfcc(X, Fs, 'maxfreq', Fs/2, 'numcep', 13, 'nbands', 14, 'fbtype', 'fcmel', 'dcttype', 1, 'usecmp', 1, 'wintime', 0.04, 'hoptime', 0.02, 'preemph', 0, 'dither', 1);
%
% Eli nidataa vastaa 13-ulotteinen MFCC joka 20:s ms.

load dt_t3_m4_mfcc.mat
[a, b] = size(mfcc);
disp(['Ladattiin ' num2str(b) ' ' num2str(a) '-D MFCC-kertointa']);

%% Lasketaan mys spektrikomponentit
% Short-time Fourier-transform coefficients a la Jukka:
% Compute spectra first and combine to 13 equal-wide bands.
winlen = round(0.04*Fs);
noverlap = round(0.02*Fs);
nfft = 24;
[XTF] = abs(specgram(X, 1300, Fs, winlen, noverlap));
for k = [1 : 13]
    stftc(k, :) = sum(XTF((k-1)*50+1: k*50, :));
end;
% uusin Matlabin versio:
%[stftc] = abs(spectrogram(X, winlen, noverlap, nfft));
disp(['Laskettiin ' num2str(size(stftc, 2)) ' ' num2str(size(stftc,1)) '-D STFTC-kerrointa']);

%% Valitaan c-means-klusteroinnin klusterin lukumr
% Anna c:lle haluttu klusterin mr.
% Mist voisi tiet oikean mrn klustereita?

c = 3;
disp(['Valittiin, ett halutaan ' num2str(c) ' klusteria.']);

%% Piirretn MFCC- ja STFTC-kertoimet

figure(42); clf;
subplot(3,1,1);
plot(t, X);
grid on;
set(gca,'XLim', [t(1) t(end)]);
title('Aaltomuoto: 10 x kiisseli');
subplot(3,1,2);
imagesc(flipud((mfcc)));
title(['Signaalista lasketut ' num2str(length(mfcc)) ' MFCC-kerrointa']);
subplot(3,1,3);
imagesc(flipud((stftc)));
title(['Signaalista lasketut ' num2str(length(stftc)) ' STFTC-kerrointa']);

%% Klusteroidaan kepstrikertoimet MFCC
% Kytetn komentoa kmeans, "help kmeans".
% Haluttu klusterin lukumr on c.

disp('Klusteroidaan MFCC-kertoimet:');
idx_mfcc = kmeans(mfcc', c, 'display','iter')';

subplot(3,1,2);
hold on;
pl1 = plot(idx_mfcc);
set(pl1, 'LineWidth', 3, 'Color', [0 0 0]);

%% Klusteroidaan Fourier-kertoimet eli STFTC
% Haluttu klusterin lukumr on c.

disp('Klusteroidaan STFTC-kertoimet:');
idx_stftc = kmeans(stftc', c, 'display','iter')';

subplot(3,1,3);
hold on;
pl2 = plot(idx_stftc);
set(pl2, 'LineWidth', 3, 'Color', [0 0 0]);

%% Kootaan tulokset nivektoreihin
% Kootuista vektoreista tulee noin tuplapituiset.

X_mfcc = cell(c, 1);
X_stftc = cell(c, 1);

for s = [1 : c]
    dummy = [];
    for m = [find(idx_mfcc == s)]
        dummy = [dummy; X(round((m-1)*441)+1 : round((m+1)*441)-1)];
    end;
    X_mfcc{s} = dummy;
end;

for s = [1 : c]
    dummy = [];
    for m = [find(idx_stftc == s)]
        dummy = [dummy; X(round((m-1)*441)+1 : round((m+1)*441)-1)];
    end;
    X_stftc{s} = dummy;
end;

%% Kuunnellaan klusterointien tulokset
% Visualisoinnin sijaan kuunnellaan milt klusterit
% kuulostavat.
%
% Miten luonnehdit syntyneit klustereita?

disp('Kuunnellaan MFCC-klusterit:');
for s = [1 : c]
    disp(['Klusteri #' num2str(s) '/' num2str(c) ' (MFCC): ' num2str(length(X_mfcc{s})/Fs) 's']);
    soundsc(X_mfcc{s}, Fs);
    disp('Paina jotain nppint jatkaaksesi...');
    pause;
end;

disp('Kuunnellaan STFTC-klusterit:');
for s = [1 : c]
    disp(['Klusteri #' num2str(s) '/' num2str(c) ' (STFTC): ' num2str(length(X_stftc{s})/Fs) 's']);
    soundsc(X_stftc{s}, Fs);
    disp('Paina jotain nppint jatkaaksesi...');
    pause;
end;

%% Huomioita
% Huomaathan, ett c-means-klusteroinnissa alkupisteet
% "arvotaan" - tm voi vaikuttaa todella paljon 
% klusterointitulokseen, joka vaihtelee "joka kerta".
% Aja siis samalla c:n arvolla useampi kerta!
